#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:

#cimport cVocalTractLabApi # Function defs could be placed here

import numpy as np
cimport numpy as np
import pandas as pd
import ctypes
import os
import warnings
import time

from libcpp cimport bool

import logging

logging.basicConfig()
log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)

import atexit

from cpython.pycapsule cimport *
#from cpython cimport array
#from libc.stdlib cimport malloc, free


from VocalTractLab.tract_sequence import Sub_Glottal_Sequence, Supra_Glottal_Sequence, Motor_Sequence
from VocalTractLab.targets import Motor_Score
from VocalTractLab.frequency_domain import Transfer_Function
from VocalTractLab.function_tools import save, load
from VocalTractLab.tube_states import Tube_State
import VocalTractLab.audio_tools as AT
import VocalTractLab.function_tools as FT



import librosa
import multiprocessing as mp
import tqdm
import itertools
from bs4 import BeautifulSoup
#import copy
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################



#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	C++ API functions:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
cdef extern from "VocalTractLabApi.h":
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	ctypedef enum SpectrumType:
		NO_RADIATION,
		PISTONINSPHERE_RADIATION,
		PISTONINWALL_RADIATION,
		PARALLEL_RADIATION,
		NUM_RADIATION_OPTIONS,
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	ctypedef enum RadiationType:
		SPECTRUM_UU,
		SPECTRUM_PU,
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	cdef struct TransferFunctionOptions:
		SpectrumType spectrumType,
		RadiationType radiationType,
		bool boundaryLayer,
		bool heatConduction,
		bool softWalls,
		bool hagenResistance,
		bool innerLengthCorrections,
		bool lumpedElements,
		bool paranasalSinuses,
		bool piriformFossa,
		bool staticPressureDrops,
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlCalcTongueRootAutomatically( bool automaticCalculation );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlClose();
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlExportTractSvg( double *tractParams,
		                   const char *fileName
		                   );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGesturalScoreToAudio( const char *gesFileName,
		                         const char *wavFileName,
		                         double *audio,
		                         int *numSamples,
		                         bool enableConsoleOutput
		                         );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGesturalScoreToTractSequence( const char *gesFileName, 
		                                 const char *tractSequenceFileName
		                                 );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetConstants( int *audioSamplingRate,
		                 int *numTubeSections,
		                 int *numVocalTractParams,
		                 int *numGlottisParams,
		                 int *numAudioSamplesPerTractState,
		                 double *internalSamplingRate
		                 );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetDefaultTransferFunctionOptions( TransferFunctionOptions *opts );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetGesturalScoreDuration( const char *gesFileName,
		                             int *numAudioSamples,
		                             int *numGestureSamples
		                             );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetGlottisParamInfo( char *names,
		                        char *descriptions,
		                        char *units,
		                        double *paramMin,
		                        double *paramMax,
		                        double *paramStandard,
		                        );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetTractParamInfo( char *names,
		                      char *descriptions,
		                      char *units,
		                      double *paramMin,
		                      double *paramMax,
		                      double *paramStandard,
		                      );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetGlottisParams( const char *shapeName,
	                         double *glottisParams,
	                         );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetTractParams( const char *shapeName,
	                       double *tractParams,
	                       );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlGetTransferFunction( double *tractParams,
	                            int numSpectrumSamples,
	                            TransferFunctionOptions *opts,
	                            double *magnitude,
	                            double *phase_rad
	                            );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	void vtlGetVersion( char *version );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlInitialize( const char *speakerFileName );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlInputTractToLimitedTract( double *inTractParams,
		                             double *outTractParams
		                             );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlSegmentSequenceToGesturalScore( const char *segFileName,
		                                   const char *gesFileName,
		                                   bool enableConsoleOutput,
		                                   );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlSynthBlock( double *tractParams,
		               double *glottisParams,
		               int numFrames,
		               int frameStep_samples,
		               double *audio,
		               bool enableConsoleOutput,
		               );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlTractSequenceToAudio( const char *tractSequenceFileName,
		                         const char *wavFileName,
		                         double *audio,
		                         int *numSamples,
		                         );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	int vtlTractToTube( double *tractParams,
	                    double *tubeLength_cm,
	                    double *tubeArea_cm2,
	                    int *tubeArticulator,
	                    double *incisorPos_cm,
	                    double *tongueTipSideElevation,
	                    double *velumOpening_cm2
	                    );
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################





#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	User functions:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
# 		Single core functions
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def automatic_calculation_of_TRX_and_TRY( bool automatic_calculation = True ):
	cdef bool automaticCalculation = automatic_calculation
	value = vtlCalcTongueRootAutomatically( automaticCalculation )
	#print( value )
	if value != 0:
		raise ValueError('VTL API function vtlCalcTongueRootAutomatically returned the Errorcode: {}  (See API doc for info.)' )
	warnings.warn( 'Automatic calculation of the Tongue Root parameters was set to {}.'.format(automatic_calculation) )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_version():
	cdef char version[32]
	vtlGetVersion( version )
	log.info( 'Compile date of the library: {}'.format( version.decode() ) )
	#if self.params.verbose == True:
	#	log.info( 'Compile date of the library: "%s"' % version.decode() )
	return version.decode()
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_constants():
	cdef int audioSamplingRate = -1
	cdef int numTubeSections = -1
	cdef int numVocalTractParams = -1
	cdef int numGlottisParams = -1
	cdef int numAudioSamplesPerTractState = -1
	cdef double internalSamplingRate = -1.0
	value = vtlGetConstants( &audioSamplingRate,
		                     &numTubeSections,
		                     &numVocalTractParams,
		                     &numGlottisParams,
		                     &numAudioSamplesPerTractState,
		                     &internalSamplingRate,
	                         )
	if value != 0:
		raise ValueError('VTL API function vtlGetConstants returned the Errorcode: {}  (See API doc for info.)'.format( value ) )
	constants = {
		'samplerate_audio': int( audioSamplingRate ),
		'samplerate_internal': float( internalSamplingRate ),
		'n_tube_sections': int( numTubeSections ),
		'n_tract_params': int( numVocalTractParams ),
		'n_glottis_params': int( numGlottisParams ),
		'n_samples_per_state': int( numAudioSamplesPerTractState ),
		}
	return constants
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_param_info( str params ):
	if params not in [ 'tract', 'glottis' ]:
		log.warning( 'Unknown key in "get_param_info". Key must be "tract" or "glottis". Returning "tract" infos now.' )
		params = 'tract'
	if params == 'tract':
		key = 'n_tract_params'
	elif params == 'glottis':
		key = 'n_glottis_params'
	constants = get_constants()
	names = ( ' ' * 10 * constants[ key ] ).encode()
	descriptions = (' ' * 100 * constants[key]).encode()
	units = (' ' * 10 * constants[key]).encode()
	cdef np.ndarray[ np.float64_t, ndim=1 ] paramMin = np.empty( constants[key], dtype='float64' )
	cdef np.ndarray[ np.float64_t, ndim=1 ] paramMax = np.empty( constants[key], dtype='float64' )
	cdef np.ndarray[ np.float64_t, ndim=1 ] paramStandard = np.empty( constants[key], dtype='float64' )
	if params == 'tract':
		value = vtlGetTractParamInfo( names, descriptions, units, &paramMin[0], &paramMax[0], &paramStandard[0] )
	elif params == 'glottis':
		value = vtlGetGlottisParamInfo( names, descriptions, units, &paramMin[0], &paramMax[0], &paramStandard[0] )
	if value != 0:
		raise ValueError('VTL API function vtlGetTractParamInfo or vtlGetGlottisParamInfo returned the Errorcode: {}  (See API doc for info.)'.format( value ) )
	descriptions = descriptions.decode().replace('\x00', '').strip(' ').strip('').split('\t')
	units = units.decode().replace('\x00', '').strip(' ').strip( '' ).split('\t')
	df = pd.DataFrame( np.array( [ descriptions, units, paramMin, paramMax, paramStandard ] ).T, columns = [ 'description', 'unit', 'min', 'max', 'standard' ] )
	df.index = names.decode().replace('\x00','').strip( ' ' ).strip('').split( '\t' )
	return df
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_shape( shape_list, str params = None, return_motor_sequence = True ):
	return get_shapes( shape_list,  params, return_motor_sequence )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_shapes( shape_list, str params = None, return_motor_sequence = True ):
	shape_list = FT.check_if_list_is_valid( shape_list, str )
	constants = get_constants()
	cdef np.ndarray[ np.float64_t, ndim=1 ] tractParams = np.empty( shape = constants[ 'n_tract_params' ],  dtype='float64' )
	cdef np.ndarray[ np.float64_t, ndim=1 ] glottisParams = np.empty( shape = constants[ 'n_glottis_params' ],  dtype='float64' )
	supra_glottal_shapes, sub_glottal_shapes = [], []
	supra_glottal_shape_names, sub_glottal_shape_names = [], []
	for shape in shape_list:
		#print(shape)
		shapeName = shape.encode()
		#if params == 'tract':
		#	value = vtlGetTractParams( shapeName, &tractParams[0, 0] )
		#	supra_glottal_shapes.append( [ tractParams, shape ] )
		#elif params == 'glottis':
		#	value = vtlGetGlottisParams( shapeName, &glottisParams[0, 0] )
		#	sub_glottal_shapes.append( [ glottisParams, shape ] )
		#else:
		value = vtlGetTractParams( shapeName, &tractParams[ 0 ] )
		if value == 0:
			supra_glottal_shapes.append( tractParams.copy() )
			supra_glottal_shape_names.append( shape )
		if value == 2:
			log.info('Specified shape: {} was not found in tract state shapes. Looking for glottal shapes now.'.format( shape ) )
			value = vtlGetGlottisParams( shapeName, &glottisParams[ 0 ] )
			if value == 0:
				sub_glottal_shapes.append( glottisParams.copy() )
				sub_glottal_shape_names.append( shape )
		if value != 0:
			raise ValueError('VTL API function vtlGetTractParams returned the Errorcode: {}  (See API doc for info.)'.format( value ) )
	supra_glottal_sequence_name = ','.join( supra_glottal_shape_names )
	sub_glottal_sequence_name = ','.join( sub_glottal_shape_names )
	#print( 'array shape: {}'.format(np.array( supra_glottal_shapes ).shape )  )
	if len( supra_glottal_shapes ) != 0 and len( sub_glottal_shapes ) != 0:
		supra_glottal_sequence = Supra_Glottal_Sequence( np.array( supra_glottal_shapes ), name = supra_glottal_sequence_name )
		sub_glottal_sequence = Sub_Glottal_Sequence( np.array( sub_glottal_shapes ), name = sub_glottal_sequence_name )
		if return_motor_sequence:
			motor_sequence_name = ','.join( [ supra_glottal_sequence_name, sub_glottal_sequence_name ] )
			return Motor_Sequence( supra_glottal_sequence, sub_glottal_sequence, name = motor_sequence_name )
		else:
			return [ supra_glottal_sequence, sub_glottal_sequence ]
	elif len( supra_glottal_shapes ) != 0:
		return Supra_Glottal_Sequence( np.array( supra_glottal_shapes ), name = supra_glottal_sequence_name )
	elif len( sub_glottal_shapes ) != 0:
		return Sub_Glottal_Sequence( np.array( sub_glottal_shapes ), name = sub_glottal_sequence_name )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_supra_glottal_state( key ):
	return get_param_info( 'tract' )[ key ].to_numpy( dtype = float )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_sub_glottal_state( key ):
	return get_param_info( 'glottis' )[ key ].to_numpy( dtype = float )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def load_speaker_file( str speaker_file_path ):
	_close()
	_initialize( speaker_file_path )
	log.info( 'Loaded new speakerfile: {}. Overwriting existing settings with the values from the new speaker file.'.format( speaker_file_path ) )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
def get_gestural_score_audio_duration( str ges_file_path, return_samples = True ):
	gesFileName = ges_file_path.encode()
	cdef int numAudioSamples = 0
	numGestureSamples = NULL
	vtlGetGesturalScoreDuration( gesFileName, &numAudioSamples, numGestureSamples )
	n_samples = numAudioSamples
	if return_samples: # returning number of audio samples
		return n_samples
	#else: # returning time in seconds
	#	return n_samples / self.params.samplerate_audio
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
# 		User mp enabled functions
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def change_gestural_score(
	in_ges_file_path_list,
	out_ges_file_path_list,
	modification_kwargs,
	workers: bool = False,
	):
	
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def gestural_score_to_audio(	ges_file_path_list,
								audio_file_path_list = None,
								save_file: bool = True,
								normalize_audio: int = None,
								sr: int = None,
								return_data: bool = False,
								workers: int = None,
								verbose: bool = False,
							):
	ges_file_path_list, audio_file_path_list = FT.check_if_input_lists_are_valid( [ ges_file_path_list, audio_file_path_list ], 
                                                                                  [ str, ( str, type(None) ) ],
	                                                                            )
	args =  [ [ges_file_path, audio_file_path, save_file, normalize_audio, sr, verbose]
		for ges_file_path, audio_file_path in itertools.zip_longest( ges_file_path_list, audio_file_path_list ) ]
	if len( args ) == 1:
		audio_data_list = _gestural_score_to_audio( args[ 0 ] )
	else:
		audio_data_list = _run_multiprocessing( _gestural_score_to_audio, args, return_data, workers )
	return audio_data_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def gestural_score_to_tract_sequence(	ges_file_path_list,  
										tract_file_path_list = None, 
										return_data: bool = False,
										workers: int = None,
									):
	ges_file_path_list, tract_file_path_list = FT.check_if_input_lists_are_valid( [ ges_file_path_list, tract_file_path_list ], 
	                                                                              [ str, ( str, type(None) ) ],
	                                                                            )
	args = [ [ges_file_path, tract_file_path, return_data ]
		for ges_file_path, tract_file_path in itertools.zip_longest( ges_file_path_list, tract_file_path_list ) ]
	tract_sequence_list = _run_multiprocessing( _gestural_score_to_tract_sequence, args, return_data, workers )
	return tract_sequence_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def segment_sequence_to_gestural_score( seg_file_path_list,
	                                    ges_file_path_list = None,
	                                    workers: int = None,
	                                    verbose: bool = False,
	                                    ):
	seg_file_path_list, ges_file_path_list = FT.check_if_input_lists_are_valid( [ seg_file_path_list, ges_file_path_list ], 
	                                                                            [ str, ( str, type(None) ) ] 
	                                                                          )
	args = [ [ seg_file_path, ges_file_path, verbose ]
		for seg_file_path, ges_file_path in itertools.zip_longest( seg_file_path_list, ges_file_path_list ) ]
	_run_multiprocessing( _segment_sequence_to_gestural_score, args, False, workers )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def tract_sequence_to_audio( motor_sequence_list,
	                         audio_file_path_list = None,
	                         save_file: bool = True,
	                         normalize_audio: int = -1,
	                         sr: int = None,
	                         return_data: bool = False,
	                         workers: int = None,
	                         verbose: bool = False,
	                         ):
	motor_sequence_list, audio_file_path_list = FT.check_if_input_lists_are_valid( [ motor_sequence_list, audio_file_path_list ], 
																		           [ ( str, Motor_Sequence, Motor_Score ),
	                                                                                 ( str, type(None) ),
	                                                                               ]
	                                                                             )
	args = [ [motor_sequence, audio_file_path, save_file, normalize_audio, sr, verbose]
		for motor_sequence, audio_file_path in itertools.zip_longest( motor_sequence_list, audio_file_path_list ) ]
	if len( args ) <= 4:
		audio_data_list = [ _tract_sequence_to_audio( arg ) for arg in args ]
	else:
		audio_data_list = _run_multiprocessing( _tract_sequence_to_audio, args, return_data, workers )
	return audio_data_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def motor_sequence_to_spectrogram(
	motor_sequence_list,
	audio_file_path_list = None,
	spectrogram_file_path_list = None,
	save_file: bool = True,
	normalize_audio: int = -1,
	sr: int = 16000,
	spectrogram_kwargs = AT.standard_16kHz_spectrogram_kwargs,
	return_data: bool = True,
	workers: int = None,
	verbose: bool = False,
	):
	motor_sequence_list, audio_file_path_list, spectrogram_file_path_list = FT.check_if_input_lists_are_valid(
		[ motor_sequence_list, audio_file_path_list, spectrogram_file_path_list ], 
		[ ( str, Motor_Sequence, Motor_Score ), ( str, type(None) ), ( str, type(None) ), ]
	)
	args = [ [
		motor_sequence,
		audio_file_path,
		spectrogram_file_path,
		save_file,
		normalize_audio,
		sr,
		spectrogram_kwargs,
		verbose,
		] for motor_sequence, audio_file_path, spectrogram_file_path in itertools.zip_longest(
			motor_sequence_list,
			audio_file_path_list,
			spectrogram_file_path_list,
		)
	]
	audio_data_list = _run_multiprocessing( _motor_sequence_to_spectrogram, args, return_data, workers )
	return audio_data_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def motor_sequence_to_melspectrogram(
	motor_sequence_list,
	audio_file_path_list = None,
	spectrogram_file_path_list = None,
	save_file: bool = True,
	normalize_audio: int = -1,
	sr: int = 16000,
	log_scale = True,
	spectrogram_kwargs = AT.standard_16kHz_spectrogram_kwargs,
	melspectrogram_kwargs = AT.standard_16kHz_melspectrogram_80_kwargs,
	return_data: bool = True,
	workers: int = None,
	verbose: bool = False,
	):
	motor_sequence_list, audio_file_path_list, spectrogram_file_path_list = FT.check_if_input_lists_are_valid(
		[ motor_sequence_list, audio_file_path_list, spectrogram_file_path_list ], 
		[ ( str, Motor_Sequence, Motor_Score ), ( str, type(None) ), ( str, type(None) ), ]
	)
	args = [ [
		motor_sequence,
		audio_file_path,
		spectrogram_file_path,
		save_file,
		normalize_audio,
		sr,
		log_scale,
		spectrogram_kwargs,
		melspectrogram_kwargs,
		verbose,
		] for motor_sequence, audio_file_path, spectrogram_file_path in itertools.zip_longest(
			motor_sequence_list,
			audio_file_path_list,
			spectrogram_file_path_list,
		)
	]
	audio_data_list = _run_multiprocessing( _motor_sequence_to_melspectrogram, args, return_data, workers )
	return audio_data_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def tract_sequence_to_limited_tract_sequence( motor_sequence,
	                                          workers: int = None, 
	                                        ):
	if not isinstance( motor_sequence, ( Motor_Sequence, Supra_Glottal_Sequence ) ):
		raise ValueError( 'motor_sequence argument must be Motor_Sequence or Supra_Glottal_Sequence, not {}'.format( type( motor_sequence ) ) )
	tract_param_data = []
	args = [ state for state in motor_sequence.states.to_numpy() ]
	tract_param_data = _run_multiprocessing( _tract_state_to_limited_tract_state, args, True, workers )
	limited_supra_glottal_sequence = Supra_Glottal_Sequence( np.array( tract_param_data ) )
	if isinstance( motor_sequence,  Motor_Sequence ):
		return Motor_Sequence( tract_states = limited_supra_glottal_sequence, glottis_states = motor_sequence.to_sub_glottal_sequence() )
	else:
		return limited_supra_glottal_sequence
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def tract_sequence_to_svg( motor_sequence_list,
	                       svg_dir_list = None,
	                       fps: int = 60,
	                       save_video = False,
	                       workers: int = None,
	                       ):
	motor_sequence_list, svg_dir_list = FT.check_if_input_lists_are_valid( [ motor_sequence_list, svg_dir_list ],
	                                                                       [ ( str, Motor_Sequence ),
	                                                                         ( str, type(None) ),
	                                                                       ]
	                                                                     )
	args = [ [motor_sequence, svg_dir, fps ]
	for motor_sequence, svg_dir in itertools.zip_longest( motor_sequence_list, svg_dir_list ) ]
	_run_multiprocessing( _tract_sequence_to_svg, args, False, workers )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def tract_sequence_to_transfer_functions( motor_sequence,
	                                      n_spectrum_samples: int = 8192,
	                                      save_magnitude_spectrum: bool = True,
	                                      save_phase_spectrum: bool = True,
	                                      workers: int = None,
	                                    ):
	if not isinstance( motor_sequence, ( Motor_Sequence, Supra_Glottal_Sequence ) ):
		raise ValueError( 'motor_sequence argument must be Motor_Sequence or Supra_Glottal_Sequence, not {}'.format( type( motor_sequence ) ) )
	if isinstance( motor_sequence, Motor_Sequence ):
		motor_sequence = motor_sequence.to_supra_glottal_sequence()
	tract_param_data = []
	args = [ [ state,
	           n_spectrum_samples,
	           save_magnitude_spectrum,
	           save_phase_spectrum ] 
		for state in motor_sequence.states.to_numpy() ]
	transfer_functions = _run_multiprocessing( _tract_state_to_transfer_function, args, True, workers )
	return transfer_functions
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def tract_sequence_to_tube_states( motor_sequence,
	                               save_tube_length: bool = True,
	                               save_tube_area: bool = True,
	                               save_tube_articulator: bool = True,
	                               save_incisor_position: bool = True,
	                               save_tongue_tip_side_elevation: bool = True,
	                               save_velum_opening: bool = True,
	                               workers: int = None,
	                             ):
	if not isinstance( motor_sequence, ( Motor_Sequence, Supra_Glottal_Sequence ) ):
		raise ValueError( 'motor_sequence argument must be Motor_Sequence or Supra_Glottal_Sequence, not {}'.format( type( motor_sequence ) ) )
	if isinstance( motor_sequence, Motor_Sequence ):
		motor_sequence = motor_sequence.to_supra_glottal_sequence()
	tract_param_data = []
	args = [ [ state,
	           save_tube_length,
	           save_tube_area,
	           save_tube_articulator,
	           save_incisor_position,
	           save_tongue_tip_side_elevation,
	           save_velum_opening ] 
		for state in motor_sequence.states.to_numpy() ]
	if len( args ) <= 4:
		tube_states = [ _tract_state_to_tube_state( arg ) for arg in args ]
	else:
		tube_states = _run_multiprocessing( _tract_state_to_tube_state, args, True, workers )
	return tube_states
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################





#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	Internal functions:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
# 		constructor / destructor functions
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _initialize( str speaker_file_path ):
	speakerFileName = speaker_file_path.encode()
	value = vtlInitialize( speakerFileName )
	if value != 0:
		raise ValueError('VTL API function vtlInitialize returned the Errorcode: {}  (See API doc for info.)'.format( value ) )
	#log.info( 'VTL API initialized.' )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _close():
	value = vtlClose()
	if value != 0:
		raise ValueError('VTL API function vtlClose returned the Errorcode: {}  (See API doc for info.)'.format( value ) )
	#log.info( 'VTL API closed.' )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
# 		multiprocessing worker functions
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#def _export_tract_svg( args ):
#	tract_state, out_file_path = args
#	cdef np.ndarray[ np.float64_t, ndim=1 ] tractParams = tract_state
#	fileName = out_file_path.encode()
#	constants = get_constants()
#	value = vtlExportTractSvg( &tractParams[0], fileName )
#	if value != 0:
#		raise ValueError('VTL API function vtlExportTractSvg returned the Errorcode: {}  (See API doc for info.)'.format( value ) )
#	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _modify_gestural_score( args ):
	in_ges_file_path, out_ges_file_path = args
	ges_file = open( in_ges_file_path ).read()
	ges_soup = BeautifulSoup( ges_file, 'html.parser' )
	#do manipulations here
	f = open( out_ges_file_path, "w" )
	f.write( ges_soup.prettify( formatter = None ) )
	f.close()
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _gestural_score_change_voice_quality( args ):
	gestural_score_soup, voice_quality_new, voice_quality_old = args
	for element in gestural_score_soup.gestural_score.find( 'gesture_sequence', {'type': 'glottal-shape-gestures'} ):
		try:
			if element.value == voice_quality_old:
				element.value = voice_quality_new
		except Exception: 
			ValueError( 'Error at element: {}, element.value: {}, vq old: {}, vq new: {}'.format( element, element.value, voice_quality_old, voice_quality_new ) )
	return gestural_score_soup
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _gestural_score_change_duration():
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _gestural_score_to_audio( args ):
	# Note that returning the number of samples via numSamples is deprecated, use getGesturalScoreAudioDuration instead!
	time_start = time.time()
	ges_file_path, audio_file_path, save_file, normalize_audio, sr, verbose = args
	constants = get_constants()
	if sr == None:
		sr = constants[ 'samplerate_audio' ]
	if not os.path.exists( ges_file_path ):
		warnings.warn( 'the specified gestural score file path does not exist: {}. API call will be skipped.'.format( ges_file_path ) )
		return
	if save_file:
		audio_file_path = FT.make_output_path( audio_file_path, ges_file_path.rsplit( '.' )[0] + '.wav' )
	if ( save_file and normalize_audio != None ) or ( save_file and sr != constants[ 'samplerate_audio' ] ):
		save_file = False
		return_audio = True
	if save_file == False:
		wavFileName = ''.encode()
	else:
		wavFileName = audio_file_path.encode()
	gesFileName = ges_file_path.encode()
	cdef np.ndarray[ np.float64_t, ndim=1 ] audio
	cdef bool enableConsoleOutput = verbose
	audio = np.zeros( get_gestural_score_audio_duration( ges_file_path, return_samples = True ), dtype='float64' )
	time_synth_start = time.time()
	cdef int numS = 0
	value = vtlGesturalScoreToAudio( gesFileName, wavFileName, &audio[0], &numS, enableConsoleOutput )
	time_synth_end = time.time()
	#print( 'elapsed synthesis time {}'.format( time_synth_end-time_synth_start ) )
	if value != 0:
		raise ValueError('VTL API function vtlGesturalScoreToAudio returned the Errorcode: {}  (See API doc for info.) \
			while processing gestural score file (input): {}, audio file (output): {}'.format(value, ges_file_path, audio_file_path) )
	if sr != constants[ 'samplerate_audio' ]:
		audio = librosa.resample( audio, constants[ 'samplerate_audio' ], sr )
	if normalize_audio != None:
		audio = AT.normalize( audio, normalize_audio )
	if verbose:
		log.info( 'Audio generated from gestural score file: {}'.format( ges_file_path ) )
	if save_file == False and audio_file_path not in ( None, '' ):
		AT.write( audio, audio_file_path, sr )
	time_end = time.time()
	#print( 'elapsed total time {}'.format( time_end-time_start ) )
	return audio
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _gestural_score_to_tract_sequence( args ):
	ges_file_path, tract_file_path, return_sequence = args
	if not os.path.exists( ges_file_path ):
		warnings.warn( 'the specified gestural score file path does not exist: {}. API call will be skipped.'.format( ges_file_path ) )
		return
	gesFileName = ges_file_path.encode()
	tract_file_path = FT.make_output_path( tract_file_path, ges_file_path.rsplit('.')[0] + '.tract' )
	tractSequenceFileName = tract_file_path.encode()
	value = vtlGesturalScoreToTractSequence( gesFileName, tractSequenceFileName )
	if value != 0:
		raise ValueError('VTL API function vtlGesturalScoreToTractSequence returned the Errorcode: {}  (See API doc for info.) \
			while processing gestural score file (input): {}, tract sequence file (output): {}'.format(value, ges_file_path, tract_file_path) )
	log.info( 'Created tractsequence file {} from gestural score file: {}'.format( tract_file_path, ges_file_path ) )
	if return_sequence:
		return Motor_Sequence.from_tract_file( tract_file_path )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _segment_sequence_to_gestural_score( args ):
	seg_file_path, ges_file_path, verbose = args
	if not os.path.exists( seg_file_path ):
		warnings.warn( 'the specified segment sequence file path does not exist: {}. API call will be skipped.'.format( seg_file_path ) )
		return
	ges_file_path = FT.make_output_path( ges_file_path, seg_file_path.rsplit('.')[0] + '.ges' )
	segFileName = seg_file_path.encode()
	gesFileName = ges_file_path.encode()
	cdef bool enableConsoleOutput = verbose
	value = vtlSegmentSequenceToGesturalScore( segFileName, gesFileName, enableConsoleOutput )
	if value != 0:
		raise ValueError('VTL API function vtlSegmentSequenceToGesturalScore returned the Errorcode: {}  (See API doc for info.) \
			while processing segment sequence file (input): {}, gestural score file (output): {}'.format(value, seg_file_path, ges_file_path) )
	log.info( 'Created gestural score from segment sequence file: {}'.format( seg_file_path ) )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _synth_block( args ):
	motor_sequence, state_samples, verbose = args
	if state_samples == None:
		constants = get_constants()
		state_samples = constants[ 'n_samples_per_state' ]
	cdef int numFrames = motor_sequence.length
	cdef np.ndarray[ np.float64_t, ndim=1 ] tractParams = motor_sequence.to_supra_glottal_states().ravel()
	cdef np.ndarray[ np.float64_t, ndim=1 ] glottisParams = motor_sequence.to_sub_glottal_states().ravel()
	cdef int frameStep_samples = state_samples
	cdef np.ndarray[ np.float64_t, ndim=1 ] audio = np.zeros( motor_sequence.length * state_samples, dtype='float64' )
	cdef bool enableConsoleOutput = verbose
	value = vtlSynthBlock( &tractParams[0], &glottisParams[0], numFrames, frameStep_samples, &audio[0], enableConsoleOutput )
	if value != 0:
		raise ValueError( 'VTL API function vtlSynthBlock returned the Errorcode: {}  (See API doc for info.)'.format( value ) )
	return audio
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _tract_sequence_to_audio( args ):
	# Note that returning the number of samples via numSamples is deprecated, use getGesturalScoreAudioDuration instead!
	motor_sequence_data, audio_file_path, save_file, normalize_audio, sr, verbose = args
	if isinstance( motor_sequence_data, str ):
		tract_file_path = motor_sequence_data
		if not os.path.exists( tract_file_path ):
			warnings.warn( 'the specified tract sequence file path does not exist: {}. API call will be skipped.'.format( tract_file_path ) )
			return
		motor_sequence = Motor_Sequence.from_tract_file( tract_file_path )
	elif isinstance( motor_sequence_data, Motor_Score ):
		motor_score = motor_sequence_data
		motor_sequence = motor_score.to_motor_sequence()
	else:
		motor_sequence = motor_sequence_data
	audio = _synth_block( ( motor_sequence, None, verbose ) )
	constants = get_constants()
	if sr == None:
		sr = constants[ 'samplerate_audio' ]
	if save_file:
		audio_file_path = FT.make_output_path( audio_file_path, motor_sequence.name.rsplit( '.' )[0] + '.wav' )
	if sr != constants[ 'samplerate_audio' ]:
		audio = librosa.resample(
			y = audio,
			orig_sr = constants[ 'samplerate_audio' ],
			target_sr = sr,
			)
	if normalize_audio != None:
		audio = AT.normalize( audio, normalize_audio )
	if save_file:
		AT.write( audio, audio_file_path, sr )
	log.info( 'Audio generated from motor_sequence: {}'.format( motor_sequence.name ) )
	return audio
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _motor_sequence_to_spectrogram( args ):
	( 
		motor_sequence_data,
		audio_file_path,
		spectrogram_file_path,
		save_file,
		normalize_audio,
		sr,
		spectrogram_kwargs,
		verbose,
	) = args
	audio_args = ( motor_sequence_data, audio_file_path, save_file, normalize_audio, sr, verbose )
	audio = _tract_sequence_to_audio( audio_args )
	spectrogram = np.abs( librosa.stft( y = audio, **spectrogram_kwargs ) )**2
	if spectrogram_file_path != None: # TODO: replace with saveFile
		spectrogram_file_path = FT.make_output_path( spectrogram_file_path, '_spectrogram.pkl.gzip' )#motor_sequence.name.rsplit( '.' )[0] +
		save( spectrogram, spectrogram_file_path )
	return spectrogram
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _motor_sequence_to_melspectrogram( args ):
	( 
		motor_sequence_data,
		audio_file_path,
		melspectrogram_file_path,
		save_file,
		normalize_audio,
		sr,
		log_scale,
		spectrogram_kwargs,
		melspectrogram_kwargs,
		verbose,
	) = args
	audio_args = ( motor_sequence_data, audio_file_path, save_file, normalize_audio, sr, verbose )
	audio = _tract_sequence_to_audio( audio_args )
	#spectrogram_kwargs = dict()
	spectrogram = np.abs( librosa.stft( y = audio, **spectrogram_kwargs ) )**2
	melspectrogram = librosa.feature.melspectrogram( S = spectrogram, **melspectrogram_kwargs )
	if log_scale == True:
		melspectrogram = librosa.power_to_db( melspectrogram )
	if melspectrogram_file_path != None: # TODO: replace with saveFile
		melspectrogram_file_path = FT.make_output_path( melspectrogram_file_path, '_spectrogram.pkl.gzip' )#motor_sequence.name.rsplit( '.' )[0] +
		save( melspectrogram, melspectrogram_file_path )
	return melspectrogram
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#def _motor_sequence_to_melspectrogram( args ):
#	( 
#		motor_sequence_data,
#		audio_file_path,
#		save_file,
#		normalize_audio,
#		sr,
#		spectrogram_file_path,
#		melspectrogram_file_path,
#		spectrogram_kwargs,
#		melspectrogram_kwargs,
#		verbose,
#	) = args
#	audio_args = ( motor_sequence_data, audio_file_path, save_file, normalize_audio, sr, verbose )
#	audio = self._tract_sequence_to_audio( audio_args )
#	spectrogram = np.abs( librosa.stft( y = audio, **spectrogram_kwargs ) )**2
#	melspectrogram = librosa.feature.melspectrogram( S = spectrogram, sr = sr, **melspectrogram_kwargs )
#	if spectrogram_file_path != None:
#		save( spectrogram, spectrogram_file_path )
#	if melspectrogram_file_path != None:
#		save( melspectrogram, melspectrogram_file_path )
#	return melspectrogram
##---------------------------------------------------------------------------------------------------------------------------------------------------#
#def _motor_sequence_to_mfcc( args ):
#	( 
#		motor_sequence_data,
#		audio_file_path,
#		save_file,
#		normalize_audio,
#		sr,
#		spectrogram_file_path,
#		melspectrogram_file_path,
#		mfcc_file_path,
#		spectrogram_kwargs,
#		melspectrogram_kwargs,
#		mfcc_kwargs,
#		verbose,
#	) = args
#	audio_args = ( motor_sequence_data, audio_file_path, save_file, normalize_audio, sr, verbose )
#	audio = self._tract_sequence_to_audio( audio_args )
#	spectrogram = np.abs( librosa.stft( y = audio, **spectrogram_kwargs ) )**2
#	melspectrogram = librosa.feature.melspectrogram( S = spectrogram, sr = sr, **melspectrogram_kwargs )
#	mfcc = librosa.feature.mfcc( S = librosa.power_to_b( dmelspectrogram ), sr = sr, **mfcc_kwargs )
#	if spectrogram_file_path != None:
#		save( spectrogram, spectrogram_file_path )
#	if melspectrogram_file_path != None:
#		save( melspectrogram, melspectrogram_file_path )
#	if mfcc_file_path != None:
#		save( mfcc, mfcc_file_path )
#	return mfcc
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _tract_state_to_limited_tract_state( args ):
	tract_state = args
	constants = get_constants()
	cdef np.ndarray[ np.float64_t, ndim=1 ] inTractParams = tract_state.ravel()
	cdef np.ndarray[ np.float64_t, ndim=1 ] outTractParams = np.zeros( constants[ 'n_tract_params' ], dtype='float64' )
	vtlInputTractToLimitedTract( &inTractParams[0], &outTractParams[0] )
	return outTractParams
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _tract_sequence_to_svg( args ):
	motor_sequence, svg_dir, fps = args
	if isinstance( motor_sequence, str ) :
		tract_file_path = motor_sequence
		if not os.path.exists( tract_file_path ) :
			warnings.warn( 'the specified tract sequence file path does not exist: {}. API call will be skipped.'.format( tract_file_path ) )
			return
		motor_sequence = Motor_Sequence.from_tract_file( tract_file_path )
	#elif isinstance( motor_sequence, ts.Target_Sequence ) :
	#	target_sequence = motor_sequence
	#	motor_sequence = target_sequence.to_motor_sequence()
	#else:
	#	#pass
	#	#motor_sequence = motor_sequence_data
	svg_dir = FT.make_output_dir( svg_dir, motor_sequence.name.rsplit('.')[0] + '_svg' )
	constants = get_constants()
	cdef np.ndarray[np.float64_t, ndim = 1] tractParams = np.zeros( constants['n_tract_params'], dtype = 'float64' )
	resampled_index = [ round(index * (44100 / 110) / fps) for index in range( 0, motor_sequence.length ) ]
	resampled_tract_states = [ [ index, tract_state ] for index, tract_state in enumerate( motor_sequence.states.to_numpy() ) if index in resampled_index ]
	for pair in resampled_tract_states:
		index, tract_state = pair
		tractParams = tract_state.ravel()
		fileName = ( svg_dir + '/tract_{}.svg'.format( index ) ).encode()
		vtlExportTractSvg( &tractParams[0], fileName )
	return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _tract_state_to_transfer_function( args ):
	tract_state, n_spectrum_samples, save_magnitude_spectrum, save_phase_spectrum = args
	magnitude_spectrum = None
	phase_spectrum = None
	cdef int numSpectrumSamples = n_spectrum_samples
	cdef np.ndarray[ np.float64_t, ndim=1 ] tractParams = tract_state.ravel()
	cdef np.ndarray[ np.float64_t, ndim=1 ] magnitude = np.zeros( n_spectrum_samples, dtype='float64' )
	cdef np.ndarray[ np.float64_t, ndim=1 ] phase_rad = np.zeros( n_spectrum_samples, dtype='float64' )
	value = vtlGetTransferFunction( &tractParams[0],
	                                numSpectrumSamples,
	                                NULL,
	                                &magnitude[0],
	                                &phase_rad[0],
	                              )
	if save_magnitude_spectrum:
		magnitude_spectrum = np.array( magnitude )
	if save_phase_spectrum:
		phase_spectrum = np.array( phase_rad )
	return Transfer_Function( magnitude_spectrum, phase_spectrum, n_spectrum_samples )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _tract_state_to_tube_state( args ):
	tract_state, save_tube_length, save_tube_area, save_tube_articulator, save_incisor_position, save_tongue_tip_side_elevation, save_velum_opening = args
	tube_length = None
	tube_area = None
	tube_articulator = None
	incisor_position = None
	tongue_tip_side_elevation = None
	velum_opening = None
	constants = get_constants()
	cdef np.ndarray[ np.float64_t, ndim=1 ] tractParams = tract_state.ravel()
	cdef np.ndarray[ np.float64_t, ndim=1 ] tubeLength_cm = np.zeros( constants[ 'n_tube_sections' ], dtype='float64' )
	cdef np.ndarray[ np.float64_t, ndim=1 ] tubeArea_cm2 = np.zeros( constants[ 'n_tube_sections' ], dtype='float64' )
	cdef np.ndarray[ int, ndim=1 ] tubeArticulator = np.zeros( constants[ 'n_tube_sections' ], dtype='i' )
	cdef double incisorPos_cm = 0.0
	cdef double tongueTipSideElevation = 0.0
	cdef double velumOpening_cm2 = 0.0
	vtlTractToTube( &tractParams[0],
	                &tubeLength_cm[0],
	                &tubeArea_cm2[0],
	                &tubeArticulator[0],
	                &incisorPos_cm,
	                &tongueTipSideElevation,
	                &velumOpening_cm2
	                )
	if save_tube_length:
		tube_length = np.array( tubeLength_cm )
	if save_tube_area:
		tube_area = np.array( tubeArea_cm2 )
	if save_tube_articulator:
		tube_articulator = np.array( tubeArticulator )
	if save_incisor_position:
		incisor_position = incisorPos_cm
	if save_tongue_tip_side_elevation:
		tongue_tip_side_elevation = tongueTipSideElevation
	if save_velum_opening:
		velum_opening = velumOpening_cm2
	return Tube_State( tube_length, tube_area, tube_articulator, incisor_position, tongue_tip_side_elevation, velum_opening )
#---------------------------------------------------------------------------------------------------------------------------------------------------#






#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _run_multiprocessing( function, args, return_data, workers ):
	if workers == None:
		workers = mp.cpu_count()
	pool = mp.Pool( workers )
	tasks = ( ( function, x ) for x in args)
	data = None
	if return_data:
		data = []
		for x in tqdm.tqdm( pool.imap( _worker, tasks ), total=len( args ) ):
			data.append( x )
	else:
		for x in tqdm.tqdm( pool.imap( _worker, tasks ), total=len( args ) ):
			pass
	pool.close()
	pool.join()
	return data
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _worker( args ):
	function, arg = args
	return function( arg )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################





#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	Code to execute at module import:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#---------------------------------------------------------------------------------------------------------------------------------------------------#
atexit.register( _close )
_initialize( os.path.join( os.path.dirname(__file__), 'speaker/JD3.speaker' ) )
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#####################################################################################################################################################


import sys,os,re
import phonecodes
import phonecode_tables

LEAF='LEAF'
MAXSET=8000

def dict_enter_word(d,w,p):
    if len(p) == 0:
        if LEAF not in d:
            d[LEAF] = set()
        d[LEAF].add(w)
    else:
        if p[0] not in d:
            d[p[0]] = {}
        dict_enter_word(d[p[0]],w,p[1:])

def dict_convert_phonecode(d0,code0,code1,language):
    d1 = dict()
    for p0 in d0:
        if p0==LEAF:
            d1[p0]=d0[p0]
        else:
            p1 = phonecodes.convert(p0,code0,code1,language)
            d1[p1] = dict_convert_phonecode(d0[p0],code0,code1,language)
    return(d1)

def dict_phones2words(d, rtd, ph, dist=0, paths=[[]], nmax=MAXSET):
    '''Return up to nmax wordstrings, in a list, that match ph
    with up to `dist' distance.'''
    res = []
    # If we're at the root node, with no ph, then current paths is a result
    # Prepend each path with the remaining dist
    if d == rtd and len(ph)==0:
        res.extend([[-dist]+x for x in paths])
    # If the next phone can extend current word, then do it
    if len(ph)>0 and ph[0] in d:
        res.extend(dict_phones2words(d[ph[0]],rtd,ph[1:],dist,paths,nmax))
    # If we've finished any word, then try to start a new one
    if d != rtd and LEAF in d:
        for w in d[LEAF]:
            ws = [ x+[w] for x in paths ]
            r=dict_phones2words(rtd,rtd,ph,dist,ws,nmax)
            while(len(res)<nmax and len(r)>0):
                res.append(r.pop())
    # If we can sub the next phone to extend current word, then do it
    if len(ph)>0 and ph[0] not in d and dist > 0:
        for p in d:
            if p != LEAF:
                res.extend(dict_phones2words(d[p],rtd,ph[1:],dist-1,paths,nmax))
    # If we can extend current word by skipping a dict phone, then do it
    if dist > 0:
        for p in d:
            if p != LEAF:
                res.extend(dict_phones2words(d[p],rtd,ph,dist-1,paths,nmax))
    # If we can extend current word by skipping a listed phone, then do it
    if len(ph)>1 and ph[1] in d and dist > 0:
        res.extend(dict_phones2words(d[ph[1]],rtd,ph[2:],dist-1,paths,nmax))
    # return up to nmax results
    res = sorted(res)
    return(res[:nmax])

###########################################################
def read_raw_dictfile(filename,language):
    '''Read from unformatted dictfile: eliminate extra spaces and parens, 
    insert a tab.  Return a list of word,pron pairs.'''
    S = []
    with open(filename) as f:
        for line in f:
            words = re.split(r'\s+',line.rstrip())
            S.append((words[0], words[1:]))
    return(S)

def read_isle_dictfile(filename,language,params):
    '''Read from ISLE-formatted dictfile, which is just IPA pronunciations,
    but word has parentheses and part of speech to be removed.
    If params['discard_phones'] is a set of chars, then any phones in the set
    will be discarded.  For example, set('#.') discards morph, syl boundaries.
    If params['discard_diacritics'] is a strings, then chars in the string
    will be discarded.  For example, 'ˈˌ' discards stress markers.
    '''
    S = []
    with open(filename) as f:
        for line in f:
            txt=re.sub(r'\([^\)]*\)','',line.rstrip())
            if 'discard_diacritics' in params:
                patstr=r'[%s]+'%(params['discard_diacritics'])
                txt = re.sub(patstr,'',txt)
            words = re.split(r'\s+',txt)
            if 'discard_phones' in params:
                pr=[p for p in words[1:] if p not in params['discard_phones']]
            else:
                pr = words[1:]
            S.append((words[0], pr))
    return(S)

_babel_pcols = {
    'amh':2,'asm':2,'ben':2,'yue':2,'ceb':1,'luo':1,'kat':2,'gug':1,'hat':1,
    'ibo':1,'jav':1,'kur':1,'lao':2,'lit':1,'mon':2,'pus':2,'swa':1,'tgl':1,
    'tam':2,'tpi':1,'tur':1,'vie':1,'zul':1
}

def read_babel_dictfile(filename, lang, params):
    '''Read from a Babel dictfile in language lang.
    If params['pcol'], then it specifies which column contains phones.
    Otherwise, try to guess based on language.
    '''
    S = []
    if 'pcol' in params:
        pcol = params['pcol']
    else:
        pcol = _babel_pcols[lang]
    with open(filename) as f:
        for line in f:
            words = re.split(r'\t',line.rstrip(), pcol)
            options = words[pcol].split('\s+')
            for option in options:
                S.append((words[0], option.split()))
    return(S)

###########################################################
_celex_pcols = { 'eng':6, 'deu':4, 'nld':4 }

def read_celex_dictfile(filename, lang, params):
    '''Read from a CELEX dictfile in language lang.'
    If params['pcol'] exists, it should be an integer,
    specifying which column contains the phonemes.
    '''
    S = []
    if 'pcol' in params:
        pcol = params['pcol']
    else:
        pcol = _celex_pcols[lang]
    with open(filename) as f:
        for line in f:
            words = re.sub(r'-','',line).rstrip().split('\\')
            word = re.sub(r'\s+','_',words[1])
            options = words[pcol].split('\\')
            for option in options:
                ol=phonecodes.attach_tones_to_vowels(list(option),set("'"),phonecode_tables._disc_vowels,1,-1)
                S.append((word, ol))
    return(S)

###################################################################
#_callhome_columns={'arz':(1,2,3,r'([@aiu%AIOUE])'),'cmn':(0,3,2,r'([iI%eEU&a@o>uR])'),'spa':(0,2,3,r'([aeiou]+)')}
_callhome_columns={'arz':(1,2,3),'cmn':(0,3,2),'spa':(0,2,3)
}
def read_callhome_dictfile(filename, L, params):
    '''Read a callhome dictionary in language L.
    If params['callhome_columns'] exists, then it should be a tuple
    (grapheme column, phone column, tone column).
    '''
    S = []
    if 'callhome_columns' in params:
        (gcol, pcol, tcol) = params['callhome_columns']
    else:
        (gcol, pcol, tcol) = _callhome_columns[L]
    with open(filename) as f:
        for line in f:
            fields = line.rstrip().split('\t')  # Split fields by tab
            tone_options = fields[tcol].split('//') # split options by slash
            syl_options = fields[pcol].split('//') 
            if len(tone_options)==1:
                options=[ (tone_options[0],s) for s in syl_options ]
            elif len(tone_options)==len(syl_options):
                options=[ (tone_options[n],syl_options[n])
                          for n in range(0,len(syl_options)) ]
            else:
                raise KeyError('tone_options and syl_options mismatched: {}'.
                               format(line.rstrip()))
            for option in options:
                syls = option[1].split(' ')
                tones = list(option[0])
                phones=[ syls[n]+tones[n] for n in range(0,len(tones)) ]
                phones.append(''.join(syls[len(tones):]))
                il = list(''.join(phones))
                ol = phonecodes.attach_tones_to_vowels(il, set(phonecode_tables._tone2ipa[L].keys()), phonecode_tables._callhome_vowels[L], -1, 1)
                S.append((fields[gcol], ol))
    return(S)

###########################################################################
def write_dictfile(w2p, filename):
    '''Write a w2p dictionary to filename'''
    lines2write = set()  # make sure each written line is unique
    for (w,pron) in w2p.items():
        if len(w)>0 and len(pron)>0:
            pl = [ re.sub(r"\s*Í¡\s*",'',p)
                   for p in pron if not p.isspace() and len(p)>0 ]
            pstr = ' '.join(pl)
            if len(pstr) > 0:
                lines2write.add('%s\t%s\n' % (w,pstr))
    with open(filename,'w') as f:
        for line in sorted(lines2write):
            f.write(line)

###################################################################
_dictread = {
    'babel':read_babel_dictfile,
    'callhome':read_callhome_dictfile,
    'celex':read_celex_dictfile,
    'isle':read_isle_dictfile
}

class lex:
    def __init__(self,language,phonecode='ipa'):
        self.p2w = {}
        self.w2p = {}
        self.language = language
        self.phonecode = phonecode
    def copy(self):
        other = lex(self.language,self.phonecode)
        other.p2w.update(self.p2w)
        other.w2p.update(self.w2p)
        return(other)
    def add(self,word,pron):
        dict_enter_word(self.p2w,word,pron)
        self.w2p[word] = pron
        return(self)
    def recode(self, code1):
        if self.phonecode==code1:
            return(self.copy())
        else:
            other = lex(self.language, code1)
            other.p2w=dict_convert_phonecode(self.p2w,self.phonecode,code1,self.language)
            for (w,p) in self.w2p.items():
                q=phonecodes.convertlist(p,self.phonecode,code1,self.language)
                other.w2p[w] = q
            return(other)
    def phones2words(self, ph, D=0, paths=[[]], nmax=MAXSET):
        '''Find all word sequences matching ph with up to D string edit dist.
        Return: a dict mapping distances to lists of lists:
        L=lex()
        res=L.phones2words(ph, 2)
        sntce in res[0] is a list of words matching ph with 0 edit distance
        sntce in res[1] is a list of words matching ph with 1 edit distance
        sntce in res[2] is a list of words matching ph with 2 edit distance
        If the total number of candidates is greater than nmax,
        then the set is truncated at nmax candidates, with a priority
        queue behavior, i.e., the highest-edit-distance candidates
        are discarded first.'''
        retval={}
        listlist=dict_phones2words(self.p2w,self.p2w,ph,D,paths,nmax)
        for res in listlist:
            if len(res)>1:
                if res[0]+D not in retval:
                    retval[res[0]+D]=[]
                retval[res[0]+D].append(res[1:])
        return(retval)
    def words2phones(self, words):
        pron = []
        for w in words:
            if w in self.w2p:
                pron.extend(self.w2p[w])
            elif w.upper() in self.w2p:
                pron.extend(self.w2p[w.upper()])
            elif w.lower() in self.w2p:
                pron.extend(self.w2p[w.lower()])
            elif w.capitalize() in self.w2p:
                pron.extend(self.w2p[w.capitalize()])
            else:
                pron.append(w)
        return(pron)
    
    def read(self,filename,lang,dicttype, params={}):
        if dicttype in _dictread:
            plist = _dictread[dicttype](filename,lang,params)
            for p in plist:
                self.add(p[0],p[1])
        else:
            plist = read_raw_dictfile(filename,lang)
            for p in plist:
                self.add(p[0],p[1])
        return(self)
    def save(self,filename):
        write_dictfile(self.w2p, filename)

_dict2phonecode = {'babel':'xsampa','celex':'disc','isle':'ipa'}
def read(filename,lang,dicttype,params={}):
    if dicttype in _dict2phonecode:
        newlex = lex(lang,_dict2phonecode[dicttype])
    else:
        newlex = lex(lang,dicttype)
    return(newlex.read(filename,lang,dicttype,params))
        



cdef extern from "VocalTractLabApi.h":
	int vtlCalcTongueRootAutomatically(bool automaticCalculation);
	void vtlGetVersion(char *version);
	int vtlInitialize(const char *speakerFileName);
	int vtlClose();
	int vtlGetConstants(int *audioSamplingRate, int *numTubeSections, int *numVocalTractParams, int *numGlottisParams);


#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
import VocalTractLab.VocalTractLabApi as vtl
from VocalTractLab.tract_sequence import Supra_Glottal_Sequence
from VocalTractLab.targets import Target
from VocalTractLab.targets import Target_Sequence
from VocalTractLab.multiprocessing_tools import _run_multiprocessing
#import matplotlib.pyplot as plt
import random
from random import getrandbits
from random import uniform
from random import randint
#from random import normal
import numpy as np
import bisect
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################



#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#
#	Generation of random target sequences:
#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def generate_target_sequences(
	n_samples,
	onset_time_range = [ 0.0, 1.0 ],
	onset_state_range = [ 0, 1 ],
	duration_range = [ 0.1, 0.4 ],
	slope_range = [ -0.75, 0.75 ],
	offset_range = [ 0, 1 ],
	tau_range = [ 0.005, 0.025 ],
	n_targets_range = [1, 20],
	balance_slope = True,
	workers = None,
	):
	args = [ [ onset_time_range, onset_state_range, duration_range, slope_range, offset_range, tau_range, str(x), n_targets_range, balance_slope ]
		for x in range( 0, n_samples) ]
	target_sequence_list = _run_multiprocessing( _generate_target_sequence, args, True, workers )
	return target_sequence_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def generate_target_contours(
	target_sequence_list,
	sr: float = 100,
	augmentation_kwargs = dict(
		resample_range = 100, # can be tuple or single float
		masking_range = [ 0.4, 0.6 ],
		random_shift_x = False,
		random_shift_y = True,
		random_masking = True,
		fixed_length = False,
		),
	workers: int = None,
	):
	args = [ [ target_sequence, sr, augmentation_kwargs ] for target_sequence in target_sequence_list ]
	contour_list = _run_multiprocessing( _generate_target_contours, args, True, workers )
	return contour_list
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def generate_supra_glottal_states(
	n_samples,
	parameter_kwargs = None,
	):
	tract_param_info = vtl.get_param_info( 'tract' )
	for parameter in parameter_kwargs:
		if parameter not in tract_param_info.index:
			raise ValueError( 'Specified parameter: {} is not a supra-glottal parameter.'.format( parameter ) )
	supra_glottal_states = []
	for parameter in tract_param_info.index:
		try:
			_min = parameter_kwargs[ parameter ][0]
			_max = parameter_kwargs[ parameter ][1]
			#print('success in dict try')
		except Exception:
			_min = float( tract_param_info.loc[ parameter, 'min' ] )
			_max = float( tract_param_info.loc[ parameter, 'max' ] )
		supra_glottal_states.append( np.random.uniform( _min, _max, n_samples ) )
	supra_glottal_sequence = Supra_Glottal_Sequence( np.array( supra_glottal_states ).T )
	try:
		for parameter, value in parameter_kwargs.items():
			supra_glottal_sequence.states[ parameter ] = value
	except AttributeError:
		pass
	return supra_glottal_sequence
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#def generate_phoneme_sequences(
#	n_samples,
#	sequence_length,
#	duration_range,
#	pitch_range,
#	):
#	for phoneme in range( 0, sequence_length ):
#		phonemes.append( vtl.phonemes()[ randint() ] )
#	segment_sequence = vtl.Segment_Sequence( phonemes )
#	gestural_score = ...
#	gestural_score.add( pitch )
#	if tract_seq:
#		ges_to_tract #inefficient
#	return





#---------------------------------------------------------------------------------------------------------------------------------------------------#
# 		single core functions
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _generate_target_sequence( args ):
	onset_time_range, onset_state_range, duration_range, slope_range, offset_range, tau_range, name, n_targets_range, balance_slope = args
	n_targets = random.randint( n_targets_range[0], n_targets_range[1] )
	onset_time = random.uniform( onset_time_range[0], onset_time_range[1] )
	if onset_state_range != None:
		onset_state = random.uniform( onset_state_range[0], onset_state_range[1] )
	else:
		onset_state = None
	#print( onset_state )
	durations = np.random.uniform( duration_range[0], duration_range[1], n_targets )
	slopes = [ 0 if ( balance_slope and getrandbits(1) ) else np.random.uniform( slope_range[0], slope_range[1] ) for _ in range(0, n_targets) ]
	offsets = np.random.uniform( offset_range[0], offset_range[1], n_targets )
	time_constants = np.random.uniform( tau_range[0], tau_range[1], n_targets )
	target_sequence = Target_Sequence( onset_time = onset_time,
	                                   durations = durations,
	                                   slopes = slopes,
	                                   offsets = offsets,
	                                   time_constants = time_constants,
	                                   name = name,
	                                   onset_state = onset_state,
	                                   )
	return target_sequence
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _generate_target_contours( args ):
	target_sequence, sr, augmentation_kwargs = args
	if augmentation_kwargs == None:
		return target_sequence.get_contour( sr )
	else:
		return _generate_augmented_target_contours( target_sequence, **augmentation_kwargs )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def _generate_augmented_target_contours(
	target_sequence,
	resample_range = 100, # can be tuple or single float
	masking_range = [ 0.4, 0.6 ],
	random_shift_x = False,
	random_shift_y = True,
	random_masking = True,
	fixed_length = False,
	):
	try:
		sr = random.uniform( resample_range[0], resample_range[1] )
	except Exception:
		sr = resample_range
	contour = target_sequence.get_contour( sr )
	if random_masking:
		for target in target_sequence.targets:
			if getrandbits(1):
				mask_percent = uniform( masking_range[0], masking_range[1] )
				mask_duration = target.duration * mask_percent
				start = uniform( target.onset_time, target.onset_time + (target.duration - mask_duration) )
				mask_start_idx = bisect.bisect_left( contour[:,0], start )
				mask_end_idx = bisect.bisect_right( contour[:,0], start + mask_duration )
				contour_x = list( contour[:,0] )
				contour_y = list( contour[:,1] )
				if fixed_length:
					contour_y[ mask_start_idx : mask_end_idx ] = [ -2 for _ in range(mask_start_idx , mask_end_idx) ]
				else:
					del contour_x[ mask_start_idx : mask_end_idx ]
					del contour_y[ mask_start_idx : mask_end_idx ]
				contour = np.array( [contour_x, contour_y] ).T
	if random_shift_x or random_shift_y:
		for index, (x, y) in enumerate( zip( contour[ :, 0 ], contour[ :, 1 ] ) ):
			if (not fixed_length) and (random_shift_x):
				contour[ index, 0 ] = x + np.random.normal( 0, 0.005 )
			if random_shift_y:
				contour[ index, 1 ] = y + np.random.normal( 0, 0.01 )
		contour[contour[:, 0].argsort()]
	return contour
#---------------------------------------------------------------------------------------------------------------------------------------------------#


#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	- This file is a part of the VocalTractLab Python module PyVTL, see https://github.com/paul-krug/VocalTractLab
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#
#	- Copyright (C) 2021, Paul Konstantin Krug, Dresden, Germany
#	- https://github.com/paul-krug/VocalTractLab
#	- Author: Paul Konstantin Krug, TU Dresden
#
#	- License info:
#
#		This program is free software: you can redistribute it and/or modify
#		it under the terms of the GNU General Public License as published by
#		the Free Software Foundation, either version 3 of the License, or
#		(at your option) any later version.
#		
#		This program is distributed in the hope that it will be useful,
#		but WITHOUT ANY WARRANTY; without even the implied warranty of
#		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#		GNU General Public License for more details.
#		
#		You should have received a copy of the GNU General Public License
#		along with this program. If not, see <http://www.gnu.org/licenses/>.
#
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#

#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
# Load essential packages:
import os
import warnings
import pandas as pd
import numpy as np
import VocalTractLab.VocalTractLabApi as vtl
import VocalTractLab.function_tools as FT
import matplotlib.pyplot as plt
from  itertools import chain
import math
import VocalTractLab.plotting_tools as PT
from VocalTractLab.plotting_tools import finalize_plot
from VocalTractLab.plotting_tools import get_plot
from VocalTractLab.plotting_tools import get_plot_limits
from VocalTractLab.plotting_tools import get_valid_tiers
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################



#####################################################################################################################################################
#---------------------------------------------------------------------------------------------------------------------------------------------------#
class State_Sequence():
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def __str__( self, ):
		return str( self.states )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def plot( self, plot_type = 'trajectory', time = 'seconds', **kwargs ):
		if plot_type in [ 'trajectory', 'trajectoris', 'time' ]:
			return self.plot_trajectories( time = time, **kwargs )
		elif plot_type in [ 'distribution', 'distributions', 'dist', 'dists' ]:
			return self.plot_distributions( **kwargs )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def plot_distributions( self, parameters = None, axs = None, n_columns = 5, plot_kwargs = PT.state_hist_kwargs, **kwargs ):
		parameters = get_valid_tiers( parameters, self.states.columns )
		n_rows = math.ceil( len( parameters ) / n_columns )
		figure, axs = get_plot( n_rows = n_rows, n_columns = n_columns, axs = axs, sharex = False, sharey = True, gridspec_kw = {} )
		index_row = 0
		index_col = 0
		for index, parameter in enumerate( parameters ):
			if index_col == n_columns:
				index_row += 1
				index_col = 0
			y = self.states.loc[ :, parameter ]
			axs[ index_row, index_col ].hist( y, **plot_kwargs.get( parameter ) )
			axs[ index_row, index_col ].set( xlabel = parameter ) #, ylim = get_plot_limits( y ) )
			index_col += 1
		#for ax in axs:
		#	ax.label_outer()
		finalize_plot( figure, axs, hide_labels = False, **kwargs )
		return axs
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def plot_trajectories( self, parameters = None, axs = None, time = 'seconds', plot_kwargs = PT.state_plot_kwargs, **kwargs ):
		parameters = PT.get_valid_tiers( parameters, self.states.columns )
		constants = vtl.get_constants()
		figure, axs = get_plot( n_rows = len( parameters ), axs = axs )
		for index, parameter in enumerate( parameters ):
			y = self.states.loc[ :, parameter ]
			x = np.array( range( 0, len( y ) ) )
			if time == 'seconds':
				x = x / constants[ 'samplerate_internal' ]
			axs[ index ].plot( x, y, **plot_kwargs.get( parameter ) )
			axs[ index ].set( ylabel = parameter, ylim = get_plot_limits( y ) )
		if time == 'seconds':
			plt.xlabel( 'Time [s]' )
		else:
			plt.xlabel( 'State' )
		#for ax in axs:
		#	ax.label_outer()
		finalize_plot( figure, axs, **kwargs )
		return axs
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################



#####################################################################################################################################################
class Sub_Glottal_Sequence( State_Sequence ):
	def __init__( self,
		          states: np.array,
		          name: str = 'sequence',
		          extension: str = '.sub_glottal',
		          ):
		self.constants = vtl.get_constants()
		if len( states.shape ) != 2:
			raise ValueError( "Shape of passed state is not two-dimensional. The shape should be (x: no. states, y: no. features)" )
		if states.shape[ 1 ] != self.constants[ 'n_glottis_params' ]:
			raise ValueError( "Dimension of features is {}, but should be {}.".format( states.shape[ 1 ], self.constants[ 'n_glottis_params' ] ) )
		self.param_info = vtl.get_param_info( 'glottis' )
		self.name = name
		self.extension = extension
		self.states = pd.DataFrame( states, columns = self.param_info.index )
		self.length = len( self.states.index )
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@classmethod
	def from_sub_glottal_sequences( cls, sub_glottal_sequence_list ):
		sub_glottal_sequence_list = FT.check_if_input_lists_are_valid( [ sub_glottal_sequence_list ], [ Sub_Glottal_Sequence ] )
		states = pd.concat( [ x.states for x in sub_glottal_sequence_list ] ).to_numpy()
		name = [ x.name for x in sub_glottal_sequence_list ] #TODO concat string in list comp
		return cls( states, name )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@classmethod
	def from_tract_file( cls, tract_file_path ):
		df_GLP = pd.read_csv( tract_file_path, delim_whitespace = True, skiprows= lambda x: read_tract_seq_GLP(x) , header = None )
		return cls( df_GLP.to_numpy(), tract_file_path )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def append( self, sub_glottal_sequence ):
		if not isinstance( sub_glottal_sequence, Sub_Glottal_Sequence ):
			raise ValueError( 'Trying to append an object of type {} to an object of type {}'.format( type( self ), type( sub_glottal_sequence ) ) )
		self.states = pd.concat( [self.states, sub_glottal_sequence.states ] )
		self.name = self.name + ',' + sub_glottal_sequence.name
		return
#####################################################################################################################################################



#####################################################################################################################################################
class Supra_Glottal_Sequence( State_Sequence ):
	def __init__( self,
		          states: np.array,
		          name: str = 'sequence',
		          extension: str = '.supra_glottal',
		          ):
		self.constants = vtl.get_constants()
		if len( states.shape ) != 2:
			raise ValueError( "Shape of passed state is not two-dimensional. The shape should be (x: no. states, y: no. features)" )
		if states.shape[ 1 ] != self.constants[ 'n_tract_params' ]:
			raise ValueError( "Dimension of features is {}, but should be {}.".format( states.shape[ 1 ], self.constants[ 'n_tract_params' ] ) )
		self.param_info = vtl.get_param_info( 'tract' )
		self.name = name
		self.extension = extension
		self.states = pd.DataFrame( states, columns = self.param_info.index )
		self.length = len( self.states.index )
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@classmethod
	def from_supra_glottal_sequences( cls, supra_glottal_sequence_list ):
		supra_glottal_sequence_list = FT.check_if_input_lists_are_valid( [ supra_glottal_sequence_list ], [ Sub_Glottal_Sequence ] )
		states = pd.concat( [ x.states for x in supra_glottal_sequence_list ] ).to_numpy()
		name = [ x.name for x in supra_glottal_sequence_list ] #TODO concat string in list comp
		return cls( states, name )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@classmethod
	def from_tract_file( cls, tract_file_path ):
		df_VTP = pd.read_csv( tract_file_path, delim_whitespace = True, skiprows= lambda x: read_tract_seq_VTP(x) , header = None )
		return cls( df_VTP.to_numpy(), tract_file_path )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def append( self, supra_glottal_sequence ):
		if not isinstance( supra_glottal_sequence, Supra_Glottal_Sequence ):
			raise ValueError( 'Trying to append a {} type object to a {} type object.'.format( type( self ), type( supra_glottal_sequence ) ) )
		self.states = pd.concat( [ self.states, supra_glottal_sequence.states ] )
		self.name = self.name + ',' + supra_glottal_sequence.name
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#	def apply_biomechanical_constraints():
#		self.tract = vtl.tract_sequence_to_limited_tract_sequence( self.tract ).tract
#####################################################################################################################################################



#####################################################################################################################################################
class Motor_Sequence( State_Sequence ):
	def __init__( self,
		          supra_glottal_sequence: Supra_Glottal_Sequence,
		          sub_glottal_sequence: Sub_Glottal_Sequence,
		          name: str = 'sequence',
		          extension: str = '.motor',
		          ):
		if not isinstance( supra_glottal_sequence, Supra_Glottal_Sequence ):
			raise TypeError( '{} type object was passed, but {} was expected.'.format( supra_glottal_sequence, Supra_Glottal_Sequence ) )
		if not isinstance( sub_glottal_sequence, Sub_Glottal_Sequence ):
			raise TypeError( '{} type object was passed, but {} was expected.'.format( sub_glottal_sequence, Sub_Glottal_Sequence ) )
		for key in supra_glottal_sequence.constants:
			if supra_glottal_sequence.constants[ key ] != sub_glottal_sequence.constants[ key ]:
				raise ValueError( 'API constant {} is different for supra and sub glottal state sequence.'.format( key ) )
		self.param_info = dict( tract = supra_glottal_sequence.param_info, glottis = sub_glottal_sequence.param_info )
		self.name = name
		self.extension = extension

		supra_glottal_states = supra_glottal_sequence.states
		sub_glottal_states = sub_glottal_sequence.states

		lengths_difference = np.abs( supra_glottal_sequence.length - sub_glottal_sequence.length )
		if supra_glottal_sequence.length > sub_glottal_sequence.length:
			warnings.warn( 'lengths of supra glottal sequence is longer than sub glottal sequence. Will pad the sub glottal sequence now.' )
			sub_glottal_states = pd.concat(
				[ sub_glottal_sequence.states, 
					pd.DataFrame( [ sub_glottal_sequence.states.iloc[ -1, : ] for _ in range(0, lengths_difference ) ] )
				], ignore_index = True )
		elif supra_glottal_sequence.length < sub_glottal_sequence.length:
			warnings.warn( 'lengths of supra glottal sequence is shorter than sub glottal sequence. Will pad the supra glottal sequence now.' )
			supra_glottal_states = pd.concat( [ supra_glottal_sequence.states, 
				                      pd.DataFrame( [ supra_glottal_sequence.states.iloc[ -1, : ] for _ in range(0, lengths_difference ) ] )
				                      ], ignore_index = True,
				                      )
		#else:
		#	supra_glottal_states = supra_glottal_sequence.states
		#	sub_glotal_states = sub_glottal_sequence.states
		if len( supra_glottal_states.index ) != len( sub_glottal_states.index ):
			raise ValueError( 'Lengths of supra- and sub-glottal parts do not match in motor_sequence: {}'.format( self.name + self.extension ) )

		self.states = pd.concat( [ supra_glottal_states, sub_glottal_states ], axis = 1 )
		self.length = len( self.states.index )
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	@classmethod
	def from_tract_file( cls, tract_file_path ):
		df_GLP = pd.read_csv( tract_file_path, delim_whitespace = True, skiprows= lambda x: read_tract_seq_GLP(x) , header = None )
		df_VTP = pd.read_csv( tract_file_path, delim_whitespace = True, skiprows= lambda x: read_tract_seq_VTP(x) , header = None )
		return cls( Supra_Glottal_Sequence( df_VTP.to_numpy() ), Sub_Glottal_Sequence( df_GLP.to_numpy() ), tract_file_path )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def append( self, motor_sequence ):
		if not isinstance( motor_sequence, Motor_Sequence ):
			raise ValueError( 'Trying to append a {} type object to a {} type object.'.format( type( self ), type( motor_sequence ) ) )
		self.states = pd.concat( [ self.states, motor_sequence.states ] )
		self.name = self.name + ',' + motor_sequence.name
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	# def apply_biomechanical_constraints( self, ):
		# self.tract = vtl.tract_sequence_to_limited_tract_sequence( self.tract ).tract
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def insert( self, parameter, trajectory, trajectory_sr = None, start = 0, time_axis = 'samples', padding = None, smooth = True ):
		if parameter not in self.states.columns:
			raise ValueError( 'The specified parameter: {} is neither a supra glottal nor a sub glottal parameter!'.format( parameter ) )
		if time_axis not in [ 'seconds', 'samples' ]:
			raise ValueError( 'Argument "time_axis" must be "seconds" or "samples", not "{}"!'.format( time_axis ) )
		trajectory = FT.check_if_list_is_valid( trajectory, (int, float) )
		state_sr = 44100/110
		if trajectory_sr != None:
			trajectory = resample_trajectory( trajectory, trajectory_sr, state_sr )
		if time_axis == 'seconds':
			start = round( state_sr * start )
		if padding == 'same':
			trajectory = [ trajectory[0] 
			               for _ in range(0, start) ] + trajectory + [ trajectory[-1] 
			               for _ in range( start + len( trajectory ), len( self.states[parameter] ) ) 
			               ] 
			#plt.plot(trajectory)
			#plt.show()
			self.states[ parameter ] = trajectory
		else:	
			self.states.loc[ start : start + len( trajectory ) - 1, parameter ] = trajectory
		#values_a = feature.loc[ : start, parameter ].to_list()
		#values_b = feature.loc[ : start, parameter ].to_list()
		#smooth_values_1 = transition( values_a, resampled_values, 40, fade='in' )
		#smooth_values_2 = transition( smooth_values_1, values_b, 40, fade='out' )
		#print(len(smooth_values_2))
		#feature.loc[ 0 : len( smooth_values_2 )-1, parameter ] = smooth_values_2
		return
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def to_sub_glottal_sequence( self ):
		return Sub_Glottal_Sequence( self.states[ self.param_info[ 'glottis' ].index ].to_numpy(), name = '' ) #TODO name
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def to_sub_glottal_states( self ):
		return self.states[ self.param_info[ 'glottis' ].index ].to_numpy()
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def to_supra_glottal_sequence( self ):
		return Supra_Glottal_Sequence( self.states[ self.param_info[ 'tract' ].index ].to_numpy(), name = '' ) #TODO name
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	def to_supra_glottal_states( self ):
		return self.states[ self.param_info[ 'tract' ].index ].to_numpy()
#---------------------------------------------------------------------------------------------------------------------------------------------------#
	# def plot( self, parameters = ['LP','JA','LD','HX','HY'], n_params = 19 ):
	# 	figure, axs = plt.subplots( len(parameters), figsize = (8, 4/3 *len(parameters) ), sharex = True, gridspec_kw = {'hspace': 0} )
	# 	#figure.suptitle( 'Sharing both axes' )
	# 	#parameters = self.tract.columns
	# 	for index, parameter in enumerate( parameters ):
	# 		axs[ index ].plot( self.tract.loc[ :, parameter ] )
	# 		axs[ index ].set( ylabel = parameter )
	# 	plt.xlabel( 'Tract state' )
	# 	for ax in axs:
	# 	    ax.label_outer()
	# 	figure.align_ylabels( axs[:] )
	# 	plt.show()
	# 	return
#####################################################################################################################################################

#---------------------------------------------------------------------------------------------------------------------------------------------------#
def resample_trajectory( trajectory, trajectory_sr, target_sr ):
	resample_rate = trajectory_sr / target_sr
	resampled_x = np.arange( 0, len(trajectory)-1, resample_rate )
	#resampled_y = interpolate( resampled_x, trajectory )
	return interpolate( resampled_x, trajectory )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def s( x, a, b ):
	return 0.5 + 0.5 * np.tanh( ( x - a ) / b )
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def h( x, g, f, a = 0, b = 2 ):
	p = s( x, a, b )
	return (p * g) + (( 1 - p ) * f)
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def transition( values_a, values_b, window_size, fade ):
	#print( 'len val a: ', len(values_a) )
	#print( 'len val b: ', len(values_b) )
	start = np.clip( len( values_a ) - int(window_size*0.5), a_min = 0, a_max = None, dtype = int )
	values_window_a = values_a[ start : ]
	values_window_b = values_b[ : window_size - len( values_window_a ) ]
	#print( 'len val w a: ', len(values_window_a) )
	#print( 'len val w b: ', len(values_window_b) )
	values_window_a = values_window_a + [ values_window_a[-1] for _ in range( 0, window_size - len(values_window_a) ) ]
	values_window_b = [ values_window_b[0] for _ in range( 0, window_size - len(values_window_b) ) ] + values_window_b
	#print( 'len val w a: ', values_window_a )
	#print( 'len val w b: ', values_window_b )
	#if fade == 'in':
	g = values_window_a
	f = values_window_b
	#elif fade == 'out':
	#	g = values_window_a
	#	f = values_window_b
	#else:
	#	raise ValueError( 'Argument "fade" must be "in" or "out", not "{}"!'.format( fade ) )
	values_window = [ h( x, f[x], g[x], window_size/2 ) for x in range( 0, len(values_window_a) ) ]
	#plt.plot( values_a[ : start ] + values_window + values_b[ window_size : ] )
	#plt.show()
	#stop
	#print( 'len val w: ', len(values_window) )
	return values_a[ : start ] + values_window + values_b[ window_size : ]
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def pad( values: list, front: int = 0, back: int = 0, type: str = 'same', smooth = True ):
	insert_front = [ value[0] for _ in range( 0, front ) ]
	insert_back = [ value[-1] for _ in range( 0, back ) ]
	return insert_front + values + insert_back
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def interpolate( values_x, values_y, type = 'linear' ):
	interpolated_values = []
	for x in values_x:
		x_0 = int( x )
		x_1 = int( x ) + 1
		y_0 = values_y[ x_0 ]
		y_1 = values_y[ x_1 ]
		interpolated_values.append( ( (y_0 * (x_1-x)) + (y_1 * (x-x_0)) ) / (x_1 - x_0) )
	return interpolated_values
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def read_tract_seq_GLP( index ):
	if (index > 7) and (index % 2 == 0):
		return False
	else:
		return True
#---------------------------------------------------------------------------------------------------------------------------------------------------#
def read_tract_seq_VTP( index ):
	if (index > 7) and ((index-1) % 2 == 0):
		return False
	else:
		return True
#---------------------------------------------------------------------------------------------------------------------------------------------------#
#####################################################################################################################################################


#!/usr/bin/env python
DOCLINES = 'This Python library provides an implementation of the articulatory synthesizer VocalTractLab.\n\
\n\
Core features are:\
- a one dimensional aero-acoustic simulation of vocal tract dynamics using a\
  vocal tract model based on MRI scans of a human vocal tract\
- synthesis of artificial speech through high-level gestural control or low-\
  level motor control of individual articulators through dynamic targets\
- extensive visualization options\
- estimation of articulatory targets from audio files directly (pitch target-\
  estimation) or from arbitrary data (e.g. articulatory measurements)\
- phoneme-to-speech functionality and a basic text-to-speech pipeline\
- and much more\
\
Besides the scientific purpose, this module is well suited for any production\
environments that need to access the VocalTractLab-backend.\
VocalTractLab and its Python module are licensed under GPL-3.0.'


import os, sys
import logging
import subprocess
import shutil

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py
from setuptools.extension import Extension

import numpy as np
from Cython.Build import cythonize
import cmake


WORKING_PATH = os.getcwd()

class Build_Target_Optimizer( build_py ):
    """Build TargetOptimizer-Backend API"""
    def run( self ):
        print( 'Building Target_Optimizer-Backend using cmake:' )
        os.chdir( 'VocalTractLab/src/targetoptimizer-backend' )
        #with TemporaryDirectory() as tmpdir:
        #    os.chdir(tmpdir)
        subprocess.check_call( [ 'cmake', '.' ] )
        subprocess.check_call( [ 'cmake', '--build', '.', '--config', 'Release' ] )
        api_name = 'TargetOptimizerApi'
        if sys.platform == 'win32':
            file_extension = '.dll'
            shutil.move( os.path.join( 'Release', api_name + file_extension ), os.path.join( WORKING_PATH, 'VocalTractLab' ) )
            shutil.move( os.path.join( 'release', api_name + '.lib' ), os.path.join( WORKING_PATH, 'VocalTractLab' ) )
        else:
            file_extension = '.so'
            shutil.move( 'lib' + api_name + file_extension, os.path.join( WORKING_PATH, 'VocalTractLab' ) )
        #print( os.listdir( os.getcwd() ) )
        #print( os.listdir( os.path.join( os.getcwd(), 'CMakeFiles' ) ) )
        shutil.move( os.path.join( '', api_name + '.h' ), os.path.join( WORKING_PATH, 'VocalTractLab' ) )
        shutil.move( os.path.join( '', 'Data.h' ), os.path.join( WORKING_PATH, 'VocalTractLab' ) )
        os.chdir( WORKING_PATH )
        #build_py.run( self )

class Build_VTL( build_py ):
    """Build VocalTractLab-Backend API"""
    def run( self ):
        print( 'Building VocalTractLab-Backend using cmake:' )
        os.chdir( 'VocalTractLab/src/vocaltractlab-backend' )
        #with TemporaryDirectory() as tmpdir:
        #    os.chdir(tmpdir)
        subprocess.check_call( [ 'cmake', '.' ] )
        subprocess.check_call( [ 'cmake', '--build', '.', '--config', 'Release' ] )
        api_name = 'VocalTractLabApi'
        if sys.platform == 'win32':
            file_extension = '.dll'
            shutil.move( os.path.join( 'Release', api_name + file_extension ), os.path.join( WORKING_PATH, 'VocalTractLab' ) )
            shutil.move( os.path.join( 'Release', api_name + '.lib' ), os.path.join( WORKING_PATH, 'VocalTractLab' ) )
        else:
            file_extension = '.so'
            shutil.move( 'lib' + api_name + file_extension, os.path.join( WORKING_PATH, 'VocalTractLab' ) )
        shutil.move( os.path.join( '', api_name + '.h' ), os.path.join( WORKING_PATH, 'VocalTractLab' ) )
        #print( ' chir dir: ' )
        #print( os.listdir( os.getcwd() ) )
        os.chdir( WORKING_PATH )
        #print( 'working dir:' )
        #print( os.listdir( os.getcwd() ) )
        #print( 'VocalTractLab dir:' )
        #print( os.listdir( os.getcwd()+'/VocalTractLab' ) )
        #build_py.run( self )

class Build_Backends( build_py ):
    def run(self):
        #self.run_command( 'build_target_optimizer' )
        self.run_command( 'build_vtl' )
        build_py.run(self)




# Set up the logging environment
logging.basicConfig()
log = logging.getLogger()

# Handle the -W all flag
if 'all' in sys.warnoptions:
    log.level = logging.DEBUG

# Get version from the VocalTractLab module
with open('VocalTractLab/__init__.py') as f:
    for line in f:
        if line.find('__version__') >= 0:
            version = line.split('=')[1].strip()
            version = version.strip('"')
            version = version.strip("'")
            continue


# Build extension modules 
EXT_MODULES = cythonize(
    [
        #Extension( 'VocalTractLab.target_estimation',
        #      ['VocalTractLab/target_estimation.pyx'],
        #      language="c++",
        #      libraries=['VocalTractLab/TargetOptimizerApi'],
        #      library_dirs=['.'],#, './src/', './src/targetoptimizer-backend/'],
        #      include_dirs=[np.get_include()],#, './src/', './src/targetoptimizer-backend/']
        #      ),
        Extension( 'VocalTractLab.VocalTractLabApi',
              ['VocalTractLab/VocalTractLabApi.pyx'],
              language="c",
              libraries=['VocalTractLab/VocalTractLabApi'],
              library_dirs=['.'],
              include_dirs=[np.get_include(), '.'],
              runtime_library_dirs=[ '.' ],
              ),
    ]
)

# Dependencies
DEPENDENCIES = [
    'librosa>=0.8.1',
    'matplotlib>=3.4.3',
    'numpy>=1.22.0',
    'pandas>=1.3.2',
    'praat-parselmouth>=0.4.0',
    'tqdm>=4.62.1',
]


CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: GNU General Public License v3 (GPLv3)
Programming Language :: C++
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: Implementation :: CPython
Topic :: Software Development
Topic :: Scientific/Engineering
Typing :: Typed
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
"""


cmdclass = dict( #build_target_optimizer = Build_Target_Optimizer,
                 build_vtl = Build_VTL,
                 build_py = Build_Backends,
                 )
#cmdclass['build_target_optimizer'] = Build_Target_Optimizer
#cmdclass['build_vtl'] = Build_VTL
#cmdclass['build_py'] = my_build

setup_args = dict(
    name='VocalTractLab',
    version=version,
    description='Articulatory (text-to-) speech synthesis for Python',
    long_description= DOCLINES,
    url='https://github.com/paul-krug/VocalTractLab',
    #download_url='https://github.com/paul-krug/VocalTractLab/archive/v_0.3.tar.gz',
    author='Paul Krug',
    author_email='paul_konstantin.krug@tu-dresden.de',
    license='GPL-3.0',
    classifiers = [_f for _f in CLASSIFIERS.split('\n') if _f],
    keywords=[ 'text-to-speech', 'speech synthesis', 'articulatory synthesis', 'vocal tract' ],
    ext_modules=EXT_MODULES,
    cmdclass = cmdclass,
    include_dirs=np.get_include(),
    packages=find_packages(),
    package_dir={'VocalTractLab': 'VocalTractLab'},
    #package_data= {'VocalTractLab': ['API/*', 'Models/*', 'Speaker/*', 'Data/*']},
    package_data= {'VocalTractLab': [ os.path.join( WORKING_PATH, 'VocalTractLab/speaker/*'),
                              os.path.join( WORKING_PATH, 'VocalTractLab/data/*'), 
                              os.path.join( WORKING_PATH, 'VocalTractLab/data/dictionaries/phonecodes/*'),
                              os.path.join( WORKING_PATH, 'VocalTractLab/src/vocaltractlab-backend/*' ),
                              os.path.join( WORKING_PATH, 'VocalTractLab/src/targetoptimizer-backend/*' ),
                              os.path.join( WORKING_PATH, 'VocalTractLab/src/targetoptimizer-backend/dlib/*' ),
                              os.path.join( WORKING_PATH,'./VocalTractLab/*' ) ]},
    include_package_data = True,
    install_requires=DEPENDENCIES,
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    #zip_safe= False,
)






setup(**setup_args)